package ru.dbotthepony.kstarbound.client.gl

import org.lwjgl.opengl.GL46.*
import org.lwjgl.system.MemoryUtil
import java.nio.ByteBuffer

enum class VBOType(val glType: Int) {
	ARRAY(GL_ARRAY_BUFFER),
	ELEMENT_ARRAY(GL_ELEMENT_ARRAY_BUFFER),
}

class VertexBufferObject(val state: GLStateTracker, val type: VBOType = VBOType.ARRAY) {
	val pointer = glGenBuffers()

	init {
		checkForGLError("Creating Vertex Buffer Object")
		state.registerCleanable(this, ::glDeleteBuffers, pointer)
	}

	val isArray get() = type == VBOType.ARRAY
	val isElementArray get() = type == VBOType.ELEMENT_ARRAY

	fun bind(): VertexBufferObject {
		state.bind(this)
		return this
	}

	fun unbind(): VertexBufferObject {
		state.unbind(this)
		return this
	}

	fun bufferData(data: ByteBuffer, usage: Int): VertexBufferObject {
		state.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: ByteBuffer, usage: Int, length: Long): VertexBufferObject {
		state.ensureSameThread()

		if (length > data.remaining().toLong()) {
			throw IndexOutOfBoundsException("Tried to upload $data into $pointer with offset at ${data.position()} with length of $length, but that is longer than remaining data length of ${data.remaining()}!")
		}

		nglNamedBufferData(pointer, length, MemoryUtil.memAddress(data), usage)

		checkForGLError()
		return this
	}

	fun bufferData(data: IntArray, usage: Int): VertexBufferObject {
		state.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: FloatArray, usage: Int): VertexBufferObject {
		state.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: DoubleArray, usage: Int): VertexBufferObject {
		state.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: LongArray, usage: Int): VertexBufferObject {
		state.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}
}
